import unittest

from sismic.code.sequence import *


class ConditionTests(unittest.TestCase):
    def setUp(self):
        self.true = SequenceCondition('True')
        self.false = SequenceCondition('False')

    def test_initial_call(self):
        self.assertEqual(self.true.evaluate(), True)
        self.assertEqual(self.false.evaluate(), False)

    def test_does_change(self):
        self.true.evaluate()
        self.false.evaluate()

        self.true.code = 'False'
        self.false.code = 'True'
        self.assertEqual(self.true.evaluate(), False)
        self.assertEqual(self.false.evaluate(), True)

    def test_forced_call(self):
        self.assertEqual(self.true.evaluate(True), True)
        self.assertEqual(self.false.evaluate(False), False)


class ConstantsTests(unittest.TestCase):
    def setUp(self):
        self.success = SequenceSuccess()
        self.failure = SequenceFailure()
        self.finish = SequenceFinish()

    def test_initial_call(self):
        self.assertEqual(self.success.evaluate(), True)
        self.assertEqual(self.failure.evaluate(), False)
        self.assertEqual(self.finish.evaluate(), False)

    def test_forced_call(self):
        self.assertEqual(self.success.evaluate(True), True)
        self.assertEqual(self.failure.evaluate(True), False)
        self.assertEqual(self.finish.evaluate(True), True)


# ################################################################


class SometimesTests(unittest.TestCase):
    def setUp(self):
        self.true = SequenceSometimes(SequenceCondition('True'))
        self.false = SequenceSometimes(SequenceCondition('False'))

    def test_initial_call(self):
        self.assertEqual(self.true.evaluate(), True)
        self.assertEqual(self.false.evaluate(), None)

    def test_true_does_not_update(self):
        self.true.evaluate()
        self.true.sequence.code = 'False'
        self.assertEqual(self.true.evaluate(), True)

    def test_false_does_update(self):
        self.false.evaluate()
        self.false.sequence.code = 'True'
        self.assertEqual(self.false.evaluate(), True)

    def test_forced_call(self):
        self.assertEqual(self.true.evaluate(True), True)
        self.assertEqual(self.false.evaluate(True), False)


class NotTests(unittest.TestCase):
    def setUp(self):
        self.ntrue = SequenceNegate(SequenceCondition('True'))
        self.nfalse = SequenceNegate(SequenceCondition('False'))
        self.nundefined = SequenceNegate(SequenceSometimes(SequenceCondition('False')))

    def test_initial_call(self):
        self.assertEqual(self.ntrue.evaluate(), False)
        self.assertEqual(self.nfalse.evaluate(), True)
        self.assertEqual(self.nundefined.evaluate(), None)

    def test_forced_call(self):
        self.assertEqual(self.ntrue.evaluate(True), False)
        self.assertEqual(self.nfalse.evaluate(True), True)
        self.assertEqual(self.nundefined.evaluate(True), True)


class AlwaysTests(unittest.TestCase):
    def setUp(self):
        self.true = SequenceCondition('True')
        self.false = SequenceCondition('False')
        self.atrue = SequenceAlways(self.true)
        self.afalse = SequenceAlways(self.false)

    def test_initial_call(self):
        self.assertEqual(self.atrue.evaluate(), None)
        self.assertEqual(self.afalse.evaluate(), False)

    def test_false_does_not_change(self):
        self.afalse.evaluate()
        self.false.code = 'True'
        self.assertEqual(self.afalse.evaluate(), False)

    def test_true_does_change(self):
        self.atrue.evaluate()
        self.true.code = 'False'
        self.assertEqual(self.atrue.evaluate(), False)

    def test_forced_call(self):
        self.assertEqual(self.atrue.evaluate(True), True)
        self.assertEqual(self.afalse.evaluate(True), False)


class NeverTests(unittest.TestCase):
    def setUp(self):
        self.true = SequenceTimedCondition(True, False)
        self.false = SequenceTimedCondition(False, True)
        self.ntrue = SequenceNever(self.true)
        self.nfalse = SequenceNever(self.false)

    def test_initial_call(self):
        self.assertEqual(self.ntrue.evaluate(), False)
        self.assertEqual(self.nfalse.evaluate(), None)

    def test_true_does_not_change(self):
        self.ntrue.evaluate()
        self.true.code = 'False'
        self.assertEqual(self.ntrue.evaluate(), False)

    def test_false_does_change(self):
        self.nfalse.evaluate()
        self.false.code = 'True'
        self.assertEqual(self.nfalse.evaluate(), False)

    def test_forced_call(self):
        self.assertEqual(self.ntrue.evaluate(True), False)
        self.assertEqual(self.nfalse.evaluate(True), True)


# ################################################################


class AndTests(unittest.TestCase):
    def setUp(self):
        self.true = SequenceCondition('True')
        self.false = SequenceCondition('False')
        self.undef = SequenceSometimes(SequenceCondition('False'))

    def test_initial_call(self):
        self.assertEqual(SequenceAnd(self.true, self.true).evaluate(), True)
        self.assertEqual(SequenceAnd(self.true, self.false).evaluate(), False)
        self.assertEqual(SequenceAnd(self.true, self.undef).evaluate(), None)
        self.assertEqual(SequenceAnd(self.false, self.true).evaluate(), False)
        self.assertEqual(SequenceAnd(self.false, self.false).evaluate(), False)
        self.assertEqual(SequenceAnd(self.false, self.undef).evaluate(), False)
        self.assertEqual(SequenceAnd(self.undef, self.true).evaluate(), None)
        self.assertEqual(SequenceAnd(self.undef, self.false).evaluate(), False)
        self.assertEqual(SequenceAnd(self.undef, self.undef).evaluate(), None)

    def test_forced_call(self):
        self.assertEqual(SequenceAnd(self.true, self.true).evaluate(True), True)
        self.assertEqual(SequenceAnd(self.true, self.false).evaluate(True), False)
        self.assertEqual(SequenceAnd(self.false, self.true).evaluate(True), False)
        self.assertEqual(SequenceAnd(self.false, self.false).evaluate(True), False)


class OrTests(unittest.TestCase):
    def setUp(self):
        self.true = SequenceCondition('True')
        self.false = SequenceCondition('False')
        self.undef = SequenceSometimes(SequenceCondition('False'))

    def test_initial_call(self):
        self.assertEqual(SequenceOr(self.true, self.true).evaluate(), True)
        self.assertEqual(SequenceOr(self.true, self.false).evaluate(), True)
        self.assertEqual(SequenceOr(self.true, self.undef).evaluate(), True)
        self.assertEqual(SequenceOr(self.false, self.true).evaluate(), True)
        self.assertEqual(SequenceOr(self.false, self.false).evaluate(), False)
        self.assertEqual(SequenceOr(self.false, self.undef).evaluate(), None)
        self.assertEqual(SequenceOr(self.undef, self.true).evaluate(), True)
        self.assertEqual(SequenceOr(self.undef, self.false).evaluate(), None)
        self.assertEqual(SequenceOr(self.undef, self.undef).evaluate(), None)

    def test_forced_call(self):
        self.assertEqual(SequenceOr(self.true, self.true).evaluate(True), True)
        self.assertEqual(SequenceOr(self.true, self.false).evaluate(True), True)
        self.assertEqual(SequenceOr(self.false, self.true).evaluate(True), True)
        self.assertEqual(SequenceOr(self.false, self.false).evaluate(True), False)


class ThenTests(unittest.TestCase):
    def setUp(self):
        self.true = SequenceCondition('True')
        self.false = SequenceCondition('False')
        self.undef = SequenceSometimes(SequenceCondition('False'))

        self.tt = SequenceThen(self.true, self.true)
        self.tf = SequenceThen(self.true, self.false)
        self.tu = SequenceThen(self.true, self.undef)
        self.ft = SequenceThen(self.false, self.true)
        self.ff = SequenceThen(self.false, self.false)
        self.fu = SequenceThen(self.false, self.undef)
        self.ut = SequenceThen(self.undef, self.true)
        self.uf = SequenceThen(self.undef, self.false)
        self.uu = SequenceThen(self.undef, self.undef)

    def test_initial_call(self):
        self.assertEqual(self.tt.evaluate(), None)
        self.assertEqual(self.tf.evaluate(), None)
        self.assertEqual(self.tu.evaluate(), None)
        self.assertEqual(self.ft.evaluate(), None)
        self.assertEqual(self.ff.evaluate(), None)
        self.assertEqual(self.fu.evaluate(), None)
        self.assertEqual(self.ut.evaluate(), None)
        self.assertEqual(self.uf.evaluate(), None)
        self.assertEqual(self.uu.evaluate(), None)

    def test_second_call(self):
        self.tt.evaluate()
        self.tf.evaluate()
        self.tu.evaluate()
        self.ft.evaluate()
        self.ff.evaluate()
        self.fu.evaluate()
        self.ut.evaluate()
        self.uf.evaluate()
        self.uu.evaluate()

        self.assertEqual(self.tt.evaluate(), True)
        self.assertEqual(self.tf.evaluate(), False)
        self.assertEqual(self.tu.evaluate(), None)
        self.assertEqual(self.ft.evaluate(), None)
        self.assertEqual(self.ff.evaluate(), None)
        self.assertEqual(self.fu.evaluate(), None)
        self.assertEqual(self.ut.evaluate(), None)
        self.assertEqual(self.uf.evaluate(), None)
        self.assertEqual(self.uu.evaluate(), None)

    def test_initial_forced_call(self):
        self.assertEqual(self.tt.evaluate(True), False)
        self.assertEqual(self.tf.evaluate(True), False)
        self.assertEqual(self.tu.evaluate(True), False)
        self.assertEqual(self.ft.evaluate(True), False)
        self.assertEqual(self.ff.evaluate(True), False)
        self.assertEqual(self.fu.evaluate(True), False)
        self.assertEqual(self.ut.evaluate(True), False)
        self.assertEqual(self.uf.evaluate(True), False)
        self.assertEqual(self.uu.evaluate(True), False)

    def test_second_forced_call(self):
        self.tt.evaluate()
        self.tf.evaluate()
        self.ft.evaluate()
        self.ff.evaluate()

        self.assertEqual(self.tt.evaluate(True), True)
        self.assertEqual(self.tf.evaluate(True), False)
        self.assertEqual(self.ft.evaluate(True), False)
        self.assertEqual(self.ff.evaluate(True), False)

    def test_delayed_condition(self):
        tf_t = SequenceThen(SequenceTimedCondition(True, False, False), self.true)
        tf_f = SequenceThen(SequenceTimedCondition(True, False, False), self.false)
        ft_t = SequenceThen(SequenceTimedCondition(False, True, True), self.true)
        ft_f = SequenceThen(SequenceTimedCondition(False, True, True), self.false)

        self.assertEqual(tf_t.evaluate(), None)
        self.assertEqual(tf_f.evaluate(), None)
        self.assertEqual(ft_t.evaluate(), None)
        self.assertEqual(ft_f.evaluate(), None)

        self.assertEqual(tf_t.evaluate(), True)
        self.assertEqual(tf_f.evaluate(), False)
        self.assertEqual(ft_t.evaluate(), None)
        self.assertEqual(ft_f.evaluate(), None)

        self.assertEqual(tf_t.evaluate(), True)
        self.assertEqual(tf_f.evaluate(), False)
        self.assertEqual(ft_t.evaluate(), True)
        self.assertEqual(ft_f.evaluate(), False)


class BuilderTests(unittest.TestCase):
    def test_constants(self):
        self.assertEqual(build_sequence('success'), SequenceSuccess())
        self.assertEqual(build_sequence('failure'), SequenceFailure())

    def test_atom(self):
        self.assertEqual(build_sequence('"True"'), SequenceSometimes(SequenceCondition('True')))
        self.assertEqual(build_sequence('never "True"'), SequenceNever(SequenceSometimes(SequenceCondition('True'))))

    def test_binary(self):
        mapping = {
            'and': SequenceAnd,
            'or': SequenceOr,
            '->': SequenceThen
        }

        for operator, func in mapping.items():
            with self.subTest(operator=operator):
                self.assertEqual(build_sequence('"True" %s "False"' % operator),
                                 func(SequenceSometimes(SequenceCondition('True')), SequenceSometimes(SequenceCondition('False'))))

    def test_operator_priority(self):
        codes = [
            ('never never success', 'never (never success)'),
            ('never success and failure', '(never success) and failure'),
            ('success and never failure', 'success and (never failure)'),

            ('success and success or success', '(success and success) or success'),
            ('success or success and success', 'success or (success and success)'),

            ('success or success -> success', '(success or success) -> success'),
            ('success -> success or success', 'success -> (success or success)'),

            ('success -> failure -> success', '(success -> failure) -> success'),
            ('success and failure and success', '(success and failure) and success'),
            ('success or failure or success', '(success or failure) or success'),
        ]

        for code_a, code_b in codes:
            with self.subTest(code_a=code_a, code_b=code_b):
                self.assertEqual(build_sequence(code_a), build_sequence(code_b))

    def test_invalid_code(self):
        codes = [
            '',
            'a',
            '"True\'',
            'success never success',
            'success andsuccess',
            'success (and success)',
            '()',
            '() and ()',
        ]
        for code in codes:
            with self.subTest(code=code):
                with self.assertRaises(StatechartError):
                    build_sequence(code)



from behave import given, then, when
from sismic.testing import steps


@given('I open the door')
@when('I open the door')
def open_the_door(context):
    return steps.send_event(context, 'door_opened')


@given('I close the door')
@when('I close the door')
def close_the_door(context):
    return steps.send_event(context, 'door_closed')


@when('I power up the microwave')
def power_up_microwave(context):
    return steps.do_nothing(context)


@given('I place an item')
@when('I place an item')
def place_an_item(context):
    return steps.send_event(context, 'item_placed')


@given('I increase the cooking duration')
def increase_cooking_duration(context):
    return steps.send_event(context, 'input_timer_inc')


@when('I press the start button')
def press_start_button(context):
    return steps.send_event(context, 'input_cooking_start')


@then('heating should be on')
def heating_is_active(context):
    return steps.state_is_active(context, 'cooking mode')


@then('heating should not be on')
def heating_not_active(context):
    return steps.state_is_not_active(context, 'cooking mode')


@then('heating should turn on')
def heating_on_is_sent(context):
    return steps.event_is_received(context, 'heating_on')


@then('heating should not turn on')
def heating_on_is_not_sent(context):
    return steps.event_is_not_received(context, 'heating_on')


@then('heating should turn off')
def heating_off_is_sent(context):
    return steps.event_is_received(context, 'heating_off')



import unittest
from unittest.mock import MagicMock

from sismic import code
from sismic.code.python import Context, FrozenContext
from sismic.exceptions import CodeEvaluationError, SequentialConditionError
from sismic.interpreter import Interpreter
from sismic.io import import_from_yaml
from sismic.model import Event, InternalEvent


class DummyEvaluatorTests(unittest.TestCase):
    def setUp(self):
        interpreter = MagicMock(name='interpreter')
        self.evaluator = code.DummyEvaluator(interpreter=interpreter)

    def test_condition(self):
        self.assertTrue(self.evaluator._evaluate_code('blablabla'))
        self.assertTrue(self.evaluator._evaluate_code('False'))

        self.assertEqual(self.evaluator.context, {})

    def test_execution(self):
        self.assertEqual(self.evaluator._execute_code('blablabla'), [])

        self.assertEqual(self.evaluator.context, {})


class FrozenContextTests(unittest.TestCase):
    def setUp(self):
        self.context = {'a': 1, 'b': 2}

    def test_freeze(self):
        freeze = FrozenContext(self.context)
        self.assertEqual(len(freeze), 2)
        self.assertEqual(freeze.a, 1)
        self.assertEqual(freeze.b, 2)

    def test_freeze_is_frozen(self):
        freeze = FrozenContext(self.context)
        self.context['a'] = 2
        self.assertEqual(freeze.a, 1)


class ContextTests(unittest.TestCase):
    def setUp(self):
        self.context = Context()

    def test_empty(self):
        self.assertEqual(0, len(self.context))

    def test_insert_root(self):
        self.context['a'] = 1
        self.assertEqual(1, self.context['a'])

    def test_delete_root(self):
        self.context['a'] = 1
        del self.context['a']
        self.assertEqual(0, len(self.context))

    def test_insert_nested(self):
        nested = self.context.new_child()
        nested['a'] = 1
        self.assertEqual(1, nested['a'])
        self.assertEqual(0, len(self.context.map))
        self.assertEqual(1, len(nested.map))

    def test_update_nested(self):
        nested_1 = self.context.new_child()
        nested_2 = self.context.new_child()

        # Nested variable is not accessible in parent or sibling
        nested_2['c'] = 1
        with self.assertRaises(KeyError):
            _ = self.context['c']
        with self.assertRaises(KeyError):
            _ = nested_1['c']
        self.assertEqual(1, nested_2['c'])

        # Variable is propagated to nested states
        self.context['a'] = 1
        self.assertEqual(1, self.context['a'])
        self.assertEqual(1, nested_1['a'])
        self.assertEqual(1, nested_2['a'])

        # Siblings have separate context
        nested_1['b'] = 2
        nested_2['b'] = 3
        self.assertEqual(2, nested_1['b'])
        self.assertEqual(3, nested_2['b'])

        # Nested contexts can affect parent context
        nested_1['a'] = 3
        self.assertEqual(3, self.context['a'])
        self.assertEqual(3, nested_2['a'])
        self.assertEqual(3, nested_2['a'])
        nested_2['a'] = 4
        self.assertEqual(4, self.context['a'])
        self.assertEqual(4, nested_1['a'])
        self.assertEqual(4, nested_2['a'])

        # Parent context is independent from nested ones
        self.context['b'] = 4
        self.assertEqual(2, nested_1['b'])
        self.assertEqual(3, nested_2['b'])
        self.assertEqual(4, self.context['b'])


class PythonEvaluatorTests(unittest.TestCase):
    def setUp(self):
        context = {
            'x': 1,
            'y': 2,
            'z': 3
        }

        interpreter = MagicMock(name='Interpreter')
        interpreter.time = 0
        interpreter.queue = MagicMock(return_value=None)
        interpreter.statechart = MagicMock()
        interpreter.configuration = []

        self.evaluator = code.PythonEvaluator(interpreter, initial_context=context)
        self.interpreter = interpreter

    def test_condition(self):
        self.assertTrue(self.evaluator._evaluate_code('True'))
        self.assertFalse(self.evaluator._evaluate_code('False'))
        self.assertTrue(self.evaluator._evaluate_code('1 == 1'))
        self.assertTrue(self.evaluator._evaluate_code('x == 1'))
        with self.assertRaises(Exception):
            self.evaluator._evaluate_code('a')

    def test_condition_on_event(self):
        self.assertTrue(self.evaluator._evaluate_code('event.data[\'a\'] == 1', additional_context={'event': Event('test', a=1)}))
        self.assertTrue(self.evaluator._evaluate_code('event.name == \'test\'', additional_context={'event': Event('test')}))

    def test_execution(self):
        self.evaluator._execute_code('a = 1')
        self.assertEqual(self.evaluator.context['a'], 1)
        self.evaluator._execute_code('x = 2')
        self.assertEqual(self.evaluator.context['x'], 2)

    def test_invalid_condition(self):
        with self.assertRaises(CodeEvaluationError):
            self.evaluator._evaluate_code('x.y')

    def test_invalid_action(self):
        with self.assertRaises(CodeEvaluationError):
            self.evaluator._execute_code('x = x.y')

    def test_send(self):
        events = self.evaluator._execute_code('send("hello")')
        self.assertEqual(events, [InternalEvent('hello')])

    def test_add_variable_in_context(self):
        self.evaluator._execute_code('a = 1\nassert a == 1', context=self.evaluator.context)
        self.assertTrue(self.evaluator._evaluate_code('a == 1', context={'a': 1}))

    @unittest.skip('http://stackoverflow.com/questions/32894942/listcomp-unable-to-access-locals-defined-in-code-called-by-exec-if-nested-in-fun')
    def test_access_outer_scope(self):
        self.evaluator._execute_code('d = [x for x in range(10) if x!=a]', context=Context({'a': 1}))


class PythonEvaluatorNestedContextTests(unittest.TestCase):
    def setUp(self):
        statechart = """
        statechart:
          name: test
          preamble: x = y = 1
          root state:
            name: root
            initial: s1
            states:
             - name: s1
               on entry: x, z = 2, 1
               transitions:
                - target: s2
                  guard: y == 1
                  action: a, z, y = 2, 2, 2
             - name: s2
        """
        sc = import_from_yaml(statechart)
        self.intp = Interpreter(sc)

    def test_initialization(self):
        self.assertEqual(self.intp.context.get('x'), 1)
        self.assertEqual(self.intp.context.get('y'), 1)
        with self.assertRaises(KeyError):
            _ = self.intp.context['z']
        with self.assertRaises(KeyError):
            _ = self.intp.context['a']

    def test_global_context(self):
        self.intp.execute()

        self.assertEqual(self.intp.context.get('x'), 2)
        self.assertEqual(self.intp.context.get('y'), 2)
        with self.assertRaises(KeyError):
            _ = self.intp.context['z']
        with self.assertRaises(KeyError):
            _ = self.intp.context['a']

    def test_nested_context(self):
        self.intp.execute()

        s1 = self.intp._evaluator.context_for('s1')
        self.assertEqual(s1['x'], 2)
        self.assertEqual(s1['y'], 2)
        self.assertEqual(s1['z'], 2)
        with self.assertRaises(KeyError):
            _ = s1['a']


class PythonEvaluatorSequenceConditionTests(unittest.TestCase):
    def setUp(self):
        self.sc = import_from_yaml("""
        statechart:
          name: test contract
          root state:
            name: root
            on entry: x = 1
            initial: s0
            states:
             - name: s0
               initial: s1
               transitions:
               - event: end
                 target: root
               states:
               - name: s1
                 transitions:
                   - target: s2
                     action: x = 2
                     event: e
               - name: s2
        """)

        self.root = self.sc.state_for('root')  # Will never be exited
        self.s0 = self.sc.state_for('s0')  # Will be exited on "end"
        self.s1 = self.sc.state_for('s1')  # Entered, and then exited on e.
        self.s2 = self.sc.state_for('s2')  # Entered when e
        self.intp = Interpreter(self.sc)

    def test_single_condition(self):
        self.root.sequences.append('"True"')
        self.intp.execute()

    def test_access_context(self):
        self.root.sequences.append('"x == 1"')
        self.intp.execute()

    def test_access_nested_context(self):
        self.s0.sequences.append('"x == 1" -> "x == 2"')
        self.intp.queue(Event('e')).queue(Event('end'))
        self.intp.execute()

    def test_fails_fast(self):
        self.s0.sequences.append('Failure')
        with self.assertRaises(SequentialConditionError):
            self.intp.execute()

    def test_fails_on_exit(self):
        self.s0.sequences.append('"x == 1" -> "x == 2" -> "x == 3"')
        self.intp.queue(Event('e'))
        self.intp.execute()
        self.intp.queue(Event('end'))
        with self.assertRaises(SequentialConditionError):
            self.intp.execute()



import argparse
import os
import shutil
import sys
import tempfile

from typing import List
from behave import __main__ as behave_main  # type: ignore


DEFAULT_STEPS_CONTENT = """
from sismic.testing.steps import *
"""

DEFAULT_ENVIRONMENT_CONTENT = """
from behave.model import Step

def before_scenario(context, scenario):
    context.execute_steps('Given I import a statechart from {{path}}')
"""

COVERAGE_ENVIRONMENT_CONTENT = """
from itertools import chain
from collections import Counter


def states_coverage(states, entered):
    entered_stats = Counter(entered)
    coverage_stat = len(entered_stats.keys()) / len(states)

    print('State coverage: {:.2%}'.format(coverage_stat))
    print('Entered states:', ' | '.join(('{} ({})'.format(k, v) for k, v in entered_stats.most_common())))
    print('Remaining states:', ' | '.join((str(s) for s in sorted(set(states).difference(entered_stats.keys())))))


def transitions_coverage(transitions, processed):
    processed_stats = Counter(processed)
    coverage_stat = len(processed_stats.keys()) / len(transitions)

    print('Transition coverage: {:.2%}'.format(coverage_stat))
    print('Processed transitions:', ' | '.join(('{} ({})'.format(k, v) for k, v in processed_stats.most_common())))


def after_scenario(context, scenario):
    trace = context._steps
    context._traces.append(trace)

    print()
    states_coverage(context._interpreter.statechart.states,
                    chain.from_iterable([step.entered_states for step in trace]))
    transitions_coverage(context._interpreter.statechart.transitions,
                         chain.from_iterable(([step.transitions for step in trace])))
    print()


def before_feature(context, feature):
    context._traces = []


def after_feature(context, feature):
    trace = list(chain.from_iterable(context._traces))
    print()
    print('Aggregated coverage data')
    print('------------------------')
    states_coverage(context._interpreter.statechart.states,
                    chain.from_iterable([step.entered_states for step in trace]))
    transitions_coverage(context._interpreter.statechart.transitions,
                         chain.from_iterable(([step.transitions for step in trace])))
    print()
"""

DEBUG_ON_ERROR = """
def after_step(context, step):
    if step.status == 'failed':
        try:
            import ipdb
            ipdb.post_mortem(step.exc_traceback)
        except ImportError:
            import pdb
            pdb.post_mortem(step.exc_traceback)
"""


def execute_behave(statechart: str, features: List[str], steps: List[str], coverage: bool, debug_on_error: bool, parameters: List[str]) -> int:
    # Create temporary directory
    with tempfile.TemporaryDirectory() as tempdir:
        # Move statechart inside
        _, statechart_filename = os.path.split(statechart)
        shutil.copy(statechart, os.path.join(tempdir, statechart_filename))

        # Move features inside
        for feature in features:
            _, feature_filename = os.path.split(feature)
            shutil.copy(feature, os.path.join(tempdir, feature_filename))

        # Create an environment file
        with open(os.path.join(tempdir, 'environment.py'), 'w') as environment:
            content = DEFAULT_ENVIRONMENT_CONTENT
            if coverage:
                content += COVERAGE_ENVIRONMENT_CONTENT
            if debug_on_error:
                content += DEBUG_ON_ERROR
            environment.write(content.replace('{{path}}', os.path.join(tempdir, statechart_filename).replace('\\', '\\\\')))

        # Create a steps subdirectory
        os.mkdir(os.path.join(tempdir, 'steps'))

        # Create steps file
        with open(os.path.join(tempdir, 'steps', '__steps.py'), 'w') as f_step:
            f_step.write(DEFAULT_STEPS_CONTENT)

        # Copy steps files
        for step in steps if steps else []:
            _, filename = os.path.split(step)
            shutil.copy(step, os.path.join(tempdir, 'steps', filename))

        # Execute behave
        # subprocess.call(['behave'] + parameters, cwd=tempdir)
        cwd = os.getcwd()
        os.chdir(tempdir)
        exit_code = behave_main.main(parameters)
        os.chdir(cwd)
        return exit_code


def main() -> int:
    parser = argparse.ArgumentParser(prog='sismic-behave',
                                     description='Command-line utility to execute Gherkin feature files using Behave.\n'
                                                 'Additional parameters will be passed to Behave.')

    parser.add_argument('statechart', metavar='statechart', type=str,
                        help='A YAML file describing a statechart')
    parser.add_argument('--features', metavar='features', nargs='+', type=str, required=True,
                        help='A list of files containing features')
    parser.add_argument('--steps', metavar='steps', nargs='+', type=str,
                        help='A list of files containing steps implementation')
    parser.add_argument('--coverage', action='store_true', default=False,
                        help='Display coverage data')
    parser.add_argument('--show-steps', action='store_true', default=False,
                        help='Display a list of available steps (equivalent to Behave\'s --steps parameter')
    parser.add_argument('--debug-on-error', action='store_true', default=False,
                        help='Drop in a debugger in case of step failure (ipdb if available)')

    args, parameters = parser.parse_known_args()
    if args.show_steps:
        parameters.append('--steps')
    return execute_behave(args.statechart, args.features, args.steps, args.coverage, args.debug_on_error, parameters)


if __name__ == '__main__':
    sys.exit(main())



import collections
import copy
from functools import partial
from itertools import chain
from types import CodeType
from typing import (Any, Callable, Dict, Iterator, List, Mapping,
                    MutableMapping, cast)

from sismic.exceptions import CodeEvaluationError
from sismic.model import (Event, InternalEvent, Statechart, StateMixin,
                          Transition)

from .evaluator import Evaluator
from .sequence import Sequence, build_sequence

__all__ = ['PythonEvaluator']


class FrozenContext(collections.Mapping):
    """
    A shallow copy of a context. The keys of the underlying context are
    exposed as attributes.
    """
    __slots__ = ['__frozencontext']

    def __init__(self, context: Mapping) -> None:
        self.__frozencontext = {k: copy.copy(v) for k, v in context.items()}

    def __getattr__(self, item):
        try:
            return self.__frozencontext[item]
        except KeyError:
            raise AttributeError('{} has no attribute {}'.format(self, item))

    def __getitem__(self, key):
        return self.__frozencontext[key]

    def __len__(self):
        return len(self.__frozencontext)

    def __iter__(self):
        return iter(self.__frozencontext)


class Context(collections.MutableMapping):
    """
    Nested context (for dealing with scopes).

    Borrowed and corrected from http://code.activestate.com/recipes/577434/

    :param data: Optional initial dict
    :param parent: Parent context, if any
    """

    __slots__ = ['parent', 'map', 'maps']

    def __init__(self, data: Mapping = None, parent: 'Context' = None) -> None:
        self.parent = parent
        self.map = cast(MutableMapping, data) if data else {}
        self.maps = [self.map]
        if parent is not None:
            self.maps += parent.maps

    def new_child(self, data: Mapping = None) -> 'Context':
        """
        Create and return a nested context.

        :param data: Optional initial dict
        :return: A nested context
        """
        return Context(data, self)

    @property
    def root(self) -> 'Context':
        """
        :return: Root context
        """
        return self if self.parent is None else self.parent.root

    def __getitem__(self, key):
        m = self.map
        for m in self.maps:
            if key in m:
                break
        return m[key]

    def __setitem__(self, key, value) -> None:
        for m in self.maps:
            if key in m:
                m[key] = value
                return
        self.map[key] = value

    def __delitem__(self, key) -> None:
        for m in self.maps:
            if key in m:
                del m[key]
                return
        del self.map[key]

    def __len__(self) -> int:
        return len(set(*chain(m.keys() for m in self.maps)))

    def __iter__(self):
        return chain.from_iterable(self.maps)

    def __contains__(self, key) -> bool:
        return any(key in m for m in self.maps)

    def __repr__(self) -> str:
        return ' -> '.join(map(repr, self.maps))


def create_send_function(event_list: List[Event]) -> Callable[..., None]:
    """
    Create and return a callable that takes a name and additional parameters, builds an InternalEvent,
    and add it into given *event_list*.

    :param event_list: list to complement
    :return: the newly created function
    """

    def send(name, **kwargs):
        event = InternalEvent(name, **kwargs)
        event_list.append(event)

    return send


class PythonEvaluator(Evaluator):
    """
    A code evaluator that understands Python.

    Depending on the method that is called, the context can expose additional values:

    - On both code execution and code evaluation:
        - A *time: float* value that represents the current time exposed by the interpreter.
        - An *active(name: str) -> bool* Boolean function that takes a state name and return *True* if and only
          if this state is currently active, ie. it is in the active configuration of the ``Interpreter`` instance
          that makes use of this evaluator.
    - On code execution:
        - A *send(name: str, **kwargs) -> None* function that takes an event name and additional keyword parameters and
          raises an internal event with it.
        - If the code is related to a transition, the *event: Event* that fires the transition is exposed.
    - On guard or contract evaluation:
        - If the code is related to a transition, the *event: Event* that fires the transition is exposed.
    - On guard or contract (except preconditions) evaluation:
        - An *after(sec: float) -> bool* Boolean function that returns *True* if and only if the source state
          was entered more than *sec* seconds ago. The time is evaluated according to Interpreter's clock.
        - A *idle(sec: float) -> bool* Boolean function that returns *True* if and only if the source state
          did not fire a transition for more than *sec* ago. The time is evaluated according to Interpreter's clock.
    - On contract (except preconditions) evaluation:
        - A variable *__old__* that has an attribute *x* for every *x* in the context when either the state
          was entered (if the condition involves a state) or the transition was processed (if the condition
          involves a transition). The value of *__old__.x* is a shallow copy of *x* at that time.
    - On contract evaluation:
        - A *sent(name: str) -> bool* function that takes an event name and return True if an event with the same name
          was sent during the current step.
        - A *received(name: str) -> bool* function  that takes an event name and return True if an event with the
          same name is currently processed in this step.

    If an exception occurred while executing or evaluating a piece of code, it is propagated by the
    evaluator.

    Each piece of code is executed with (a partially isolated) local context.
    Every state and every transition has a specific execution context.
    The code associated with a state is executed in a local context which is composed of local variables and every
    variable that is defined in the context of the parent state (and so one until the root context is reached).
    The context of a transition is built upon the context of its source state.
    The specific context of a state is available through the *context_for* method of a PythonEvaluator.

    :param interpreter: the interpreter that will use this evaluator,
        is expected to be an *Interpreter* instance
    :param initial_context: a dictionary that will be used as *__locals__*
    """

    def __init__(self, interpreter=None, *, initial_context: Mapping[str, Any]=None) -> None:
        super().__init__(interpreter, initial_context=initial_context)

        self._context = Context(initial_context)
        self._interpreter = interpreter

        # Memory and entry time
        self._memory = {}  # type: Dict[int, Mapping]
        self._entry_time = {}  # type: Dict[str, float]
        self._idle_time = {}  # type: Dict[str, float]

        # Precompiled code
        self._evaluable_code = {}  # type: Dict[str, CodeType]
        self._executable_code = {}  # type: Dict[str, CodeType]

        # Contexts
        self._contexts = {}  # type: Dict[str, Context]
        if getattr(interpreter, 'statechart', None) is not None:
            # Initialize nested contexts
            sc = self._interpreter.statechart  # type: Statechart

            self._contexts[sc.root] = self._context.new_child()
            for name in sc.descendants_for(sc.root):
                parent_name = sc.parent_for(name)
                self._contexts[name] = self._contexts[parent_name].new_child()

        # Intercept sent and received events
        self._sents_events = []  # type: List[Event]
        if self._interpreter is not None:
            self._interpreter.bind(self._sents_events.append)
        self._received_event = None  # type: Event

    @property
    def context(self) -> Mapping:
        context = dict(self._context)
        for state, mapping in self._contexts.items():
            for key, value in mapping.map.items():
                context['{}.{}'.format(state, key)] = value
        return context

    def on_step_starts(self, event: Event = None) -> None:
        """
        Called each time the interpreter starts a macro step.

        :param event: Optional processed event
        """
        self._sents_events.clear()
        self._received_event = event

    def context_for(self, name: str) -> Context:
        """
        Context object for given state name.

        :param name: State name
        :return: Context object
        """
        return self._contexts[name]

    def _received(self, name: str) -> bool:
        """
        :param name: name of an event
        :return: True if given event name was received in current step.
        """
        return getattr(self._received_event, 'name', None) == name

    def _sent(self, name: str) -> bool:
        """
        :param name: name of an event
        :return: True if given event name was sent during this step.
        """
        return any((name == e.name for e in self._sents_events))

    def _active(self, name: str) -> bool:
        """
        Return True if given state name is active.

        :param name: name of a state
        :return: True if given state name is active.
        """
        return name in self._interpreter.configuration

    def _after(self, name: str, seconds: float) -> bool:
        """
        Return True if given state was entered more than *seconds* ago.

        :param name: name of a state
        :param seconds: elapsed time
        :return: True if given state was entered more than *seconds* ago.
        """
        return self._interpreter.time - seconds >= self._entry_time[name]

    def _idle(self, name: str, seconds: float) -> bool:
        """
        Return True if given state was the target of a transition more than *seconds* ago.

        :param name: name of a state
        :param seconds: elapsed time
        :return: True if given state was the target of a transition more than *seconds* ago.
        """
        return self._interpreter.time - seconds >= self._idle_time[name]

    def _evaluate_code(self, code: str, *, additional_context: Mapping = None, context: Mapping = None) -> bool:
        """
        Evaluate given code using Python.

        :param code: code to evaluate
        :param additional_context: an optional additional context
        :param context: the current context for this piece of code
        :return: truth value of *code*
        """
        if not code:
            return True
        if context is None:
            context = self._context

        compiled_code = self._evaluable_code.get(code, None)
        if compiled_code is None:
            compiled_code = self._evaluable_code.setdefault(code, compile(code, '<string>', 'eval'))

        exposed_context = {
            'active': self._active,
            'time': self._interpreter.time,
        }
        exposed_context.update(additional_context if additional_context is not None else {})

        try:
            return eval(compiled_code, exposed_context, context)  # type: ignore
        except Exception as e:
            raise CodeEvaluationError('The above exception occurred while evaluating:\n{}'.format(code)) from e

    def _execute_code(self, code: str, *, additional_context: Mapping = None, context: Mapping = None) -> List[Event]:
        """
        Execute given code using Python.

        :param code: code to execute
        :param additional_context: an optional additional context
        :param context: the current context for this piece of code
        :return: a list of sent events
        """
        if not code:
            return []
        if context is None:
            context = self._context

        compiled_code = self._executable_code.get(code, None)
        if compiled_code is None:
            compiled_code = self._executable_code.setdefault(code, compile(code, '<string>', 'exec'))

        # Handle sent events
        sent_events = []  # type: List[Event]

        exposed_context = {
            'active': self._active,
            'send': create_send_function(sent_events),
            'time': self._interpreter.time,
        }
        exposed_context.update(additional_context if additional_context is not None else {})

        try:
            exec(compiled_code, exposed_context, context)  # type: ignore
            return sent_events
        except Exception as e:
            raise CodeEvaluationError('The above exception occurred while executing:\n{}'.format(code)) from e

    def execute_statechart(self, statechart: Statechart) -> List[Event]:
        """
        Execute the initial code of a statechart.
        This method is called at the very beginning of the execution.

        :param statechart: statechart to consider
        :return: a list of sent events
        """
        if statechart.preamble:
            return self._execute_code(statechart.preamble, context=self._context)
        else:
            return []

    def evaluate_guard(self, transition: Transition, event: Event) -> bool:
        """
        Evaluate the guard for given transition.

        :param transition: the considered transition
        :param event: instance of *Event* if any
        :return: truth value of *code*
        """
        additional_context = {
            'event': event,
            'after': partial(self._after, transition.source),
            'idle': partial(self._idle, transition.source),
        }
        return self._evaluate_code(getattr(transition, 'guard', None),
                                   context=self._contexts[transition.source].new_child(),
                                   additional_context=additional_context)

    def execute_action(self, transition: Transition, event: Event) -> List[Event]:
        """
        Execute the action for given transition.
        This method is called for every transition that is processed, even those with no *action*.

        :param transition: the considered transition
        :param event: instance of *Event* if any
        :return: a list of sent events
        """
        self._idle_time[transition.source] = self._interpreter.time

        return self._execute_code(getattr(transition, 'action', None),
                                  context=self._contexts[transition.source].new_child(),
                                  additional_context={'event': event})

    def execute_onentry(self, state: StateMixin) -> List[Event]:
        """
        Execute the on entry action for given state.
        This method is called for every state that is entered, even those with no *on_entry*.

        :param state: the considered state
        :return: a list of sent events
        """
        self._entry_time[state.name] = self._interpreter.time
        self._idle_time[state.name] = self._interpreter.time

        return self._execute_code(getattr(state, 'on_entry', None), context=self._contexts[state.name])

    def execute_onexit(self, state: StateMixin) -> List[Event]:
        """
        Execute the on exit action for given state.
        This method is called for every state that is exited, even those with no *on_exit*.

        :param state: the considered state
        :return: a list of sent events
        """
        return self._execute_code(getattr(state, 'on_exit', None), context=self._contexts[state.name])

    def evaluate_preconditions(self, obj, event: Event = None) -> Iterator[str]:
        """
        Evaluate the preconditions for given object (either a *StateMixin* or a
        *Transition*) and return a list of conditions that are not satisfied.

        :param obj: the considered state or transition
        :param event: an optional *Event* instance, in the case of a transition
        :return: list of unsatisfied conditions
        """
        state_name = obj.source if isinstance(obj, Transition) else obj.name
        context = self._contexts[state_name]

        additional_context = {'event': event} if isinstance(obj, Transition) else {}
        additional_context.update({
            'received': self._received,
            'sent': self._sent
        })

        # Only needed if there is an invariant, a postcondition or a sequential condition
        if len(getattr(obj, 'invariants', [])) > 0 or len(getattr(obj, 'postconditions', [])) > 0 or len(
                getattr(obj, 'sequences', [])) > 0:
            self._memory[id(obj)] = FrozenContext(context)

        return filter(
            lambda c: not self._evaluate_code(c, context=context, additional_context=additional_context),
            getattr(obj, 'preconditions', [])
        )

    def evaluate_invariants(self, obj, event: Event = None) -> Iterator[str]:
        """
        Evaluate the invariants for given object (either a *StateMixin* or a
        *Transition*) and return a list of conditions that are not satisfied.

        :param obj: the considered state or transition
        :param event: an optional *Event* instance, in the case of a transition
        :return: list of unsatisfied conditions
        """
        state_name = obj.source if isinstance(obj, Transition) else obj.name
        context = self._contexts[state_name]

        additional_context = {'event': event} if isinstance(obj, Transition) else {}  # type: Dict[str, Any]
        additional_context.update({
            '__old__': self._memory.get(id(obj), None),
            'after': partial(self._after, state_name),
            'idle': partial(self._idle, state_name),
            'received': self._received,
            'sent': self._sent,
        })

        return filter(
            lambda c: not self._evaluate_code(c, context=context, additional_context=additional_context),
            getattr(obj, 'invariants', [])
        )

    def evaluate_postconditions(self, obj, event: Event = None) -> Iterator[str]:
        """
        Evaluate the postconditions for given object (either a *StateMixin* or a
        *Transition*) and return a list of conditions that are not satisfied.

        :param obj: the considered state or transition
        :param event: an optional *Event* instance, in the case of a transition
        :return: list of unsatisfied conditions
        """
        state_name = obj.source if isinstance(obj, Transition) else obj.name
        context = self._contexts[state_name]

        additional_context = {'event': event} if isinstance(obj, Transition) else {}  # type: Dict[str, Any]
        additional_context.update({
            '__old__': self._memory.get(id(obj), None),
            'after': partial(self._after, state_name),
            'idle': partial(self._idle, state_name),
            'received': self._received,
            'sent': self._sent,
        })

        return filter(
            lambda c: not self._evaluate_code(c, context=context, additional_context=additional_context),
            getattr(obj, 'postconditions', [])
        )

    def _evaluate_sequential_conditions_for_state(self, state: StateMixin, code: str) -> bool:
        context = self._contexts[state.name]

        additional_context = {
            '__old__': self._memory.get(id(state), None),
            'after': partial(self._after, state.name),
            'idle': partial(self._idle, state.name),
            'received': self._received,
            'sent': self._sent,
        }
        return self._evaluate_code(code, context=context, additional_context=additional_context)

    def initialize_sequential_conditions(self, state: StateMixin) -> None:
        """
        Initialize sequential conditions.

        :param state: for given state.
        """
        condition_mapping = {}  # type: Dict[str, Sequence]
        func = cast(Callable[[str], bool], partial(self._evaluate_sequential_conditions_for_state, state))

        for condition in getattr(state, 'sequences', []):
            condition_mapping[condition] = build_sequence(condition, func)

        self._condition_sequences[state.name] = condition_mapping



import os
import unittest

from schema import SchemaError
from sismic import exceptions, io


class ImportFromYamlParserTests(unittest.TestCase):
    def check_type(self, name):
        yaml = ('statechart:'
                '\n  name: ' + str(name) +
                '\n  preamble: Nothing'
                '\n  root state:'
                '\n    name: s1')

        item = io.import_from_yaml(yaml).name
        self.assertIsInstance(item, str, msg=type(item))

    def test_parser_conversion(self):
        self.check_type(1)
        self.check_type(-1)
        self.check_type(1.0)
        self.check_type('yes')
        self.check_type('True')
        self.check_type('no')
        self.check_type('')

        self.check_type([])
        self.check_type([1, 2])
        self.check_type({})
        self.check_type({1: 1})


class ImportFromYamlTests(unittest.TestCase):
    def test_yaml_tests(self):
        files = ['actions', 'composite', 'deep_history', 'infinite', 'internal', 'nested_parallel',
                 'nondeterministic', 'parallel', 'simple', 'timer']
        for filename in files:
            with self.subTest(filename=filename):
                with open(os.path.join('tests', 'yaml', filename + '.yaml')) as f:
                    io.import_from_yaml(f)

    def test_examples(self):
        files = ['elevator/elevator', 'elevator/elevator_contract', 'microwave/microwave',
                 'elevator/tester_elevator_7th_floor_never_reached', 'elevator/tester_elevator_moves_after_10s', 'writer_options']
        for filename in files:
            with self.subTest(filename=filename):
                with open(os.path.join('docs', 'examples', filename + '.yaml')) as f:
                    io.import_from_yaml(f)

    def test_transitions_to_unknown_state(self):
        yaml = """
        statechart:
          name: test
          root state:
            name: root
            initial: s1
            states:
              - name: s1
                transitions:
                  - target: s2
        """
        with self.assertRaises(exceptions.StatechartError) as cm:
            io.import_from_yaml(yaml)
        self.assertIn('Unknown target state', str(cm.exception))

    def test_history_not_in_compound(self):
        yaml = """
        statechart:
          name: test
          root state:
            name: root
            initial: s1
            states:
              - name: s1
                parallel states:
                 - name: s2
                   type: shallow history
        """
        with self.assertRaises(exceptions.StatechartError) as cm:
            io.import_from_yaml(yaml)
        self.assertIn('cannot be used as a parent for', str(cm.exception))

    def test_declare_both_states_and_parallel_states(self):
        yaml = """
        statechart:
          name: test
          root state:
            name: root
            initial: s1
            states:
              - name: s1
            parallel states:
              - name: s2
        """

        with self.assertRaises(exceptions.StatechartError) as cm:
            io.import_from_yaml(yaml)
        self.assertIn('root cannot declare both a "states" and a "parallel states" property', str(cm.exception))


class ExportToDictYAMLTests(unittest.TestCase):
    def setUp(self):
        files_t = ['actions', 'composite', 'deep_history', 'infinite', 'internal', 'nested_parallel',
                   'nondeterministic', 'parallel', 'simple', 'timer']
        files_e = ['elevator/elevator', 'elevator/elevator_contract', 'microwave/microwave',
                   'elevator/tester_elevator_7th_floor_never_reached', 'elevator/tester_elevator_moves_after_10s', 'writer_options']

        self.files = []
        for filename in files_t:
            with open(os.path.join('tests', 'yaml', filename + '.yaml')) as f:
                self.files.append(('\n'.join(f.readlines()), filename))
        for filename in files_e:
            with open(os.path.join('docs', 'examples', filename + '.yaml')) as f:
                self.files.append(('\n'.join(f.readlines()), filename))

    def test_export(self):
        for f, filename in self.files:
            with self.subTest(filename=filename):
                sc_1 = io.import_from_yaml(f)
                io.export_to_yaml(sc_1)

    def test_export_valid(self):
        for f, filename in self.files:
            with self.subTest(filename=filename):
                sc_1 = io.import_from_yaml(f)
                ex_1 = io.export_to_yaml(sc_1)
                sc_2 = io.import_from_yaml(ex_1)

                self.assertTrue(sc_2.validate())


class ExportToTreeTests(unittest.TestCase):
    def setUp(self):
        files_t = ['actions', 'composite', 'deep_history', 'infinite', 'internal', 'nested_parallel',
                   'nondeterministic', 'parallel', 'simple', 'timer']

        self.files = []
        for filename in files_t:
            with open(os.path.join('tests', 'yaml', filename + '.yaml')) as f:
                self.files.append((filename, io.import_from_yaml(f)))

    def test_output(self):
        results = [
            ['root', '   s1', '   s2', '   s3'],
            ['root', '   s1', '      s1a', '      s1b', '         s1b1', '         s1b2', '   s2'],
            ['root', '   active', '      active.H*', '      concurrent_processes', '         process_1',
             '            s11', '            s12', '            s13', '         process_2', '            s21',
             '            s22', '            s23', '   pause', '   stop'],
            ['root', '   s1', '   s2', '   stop'],
            ['root', '   active', '   s1', '   s2'],
            ['root', '   s1', '      p1', '         r1', '            i1', '            j1', '            k1',
             '         r2', '            i2', '            j2', '            k2', '            y', '      p2',
             '         r3', '            i3', '            j3', '            k3', '            z', '         r4',
             '            i4', '            j4', '            k4', '            x'],
            ['root', '   s1', '   s2', '   s3'],
            ['root', '   s1', '      p1', '         a1', '         b1', '         c1', '         initial1', '      p2',
             '         a2', '         b2', '         c2', '         initial2'],
            ['root', '   final', '   s1', '   s2', '   s3'],
            ['root', '   s1', '   s2', '   s3', '   s4']
        ]

        for (filename, statechart), r in zip(self.files, results):
            with self.subTest(filename=filename):
                self.assertEqual(io.text.export_to_tree(statechart), r)

    def test_all_states_are_exported(self):
        for filename, statechart in self.files:
            with self.subTest(filename=filename):
                result = set(io.text.export_to_tree(statechart, spaces=0))
                expected = set(statechart.states)
                self.assertSetEqual(result, expected)

    def test_indentation_is_correct(self):
        for filename, statechart in self.files:
            with self.subTest(filename=filename):
                result = sorted(io.text.export_to_tree(statechart, spaces=1))

                for r in result:
                    name = r.lstrip()
                    depth = statechart.depth_for(name)
                    spaces = len(r) - len(name)
                    self.assertEqual(depth - 1, spaces)



import unittest

from sismic import exceptions, io, model


class EventTests(unittest.TestCase):
    def test_creation(self):
        self.assertEqual(model.Event(name='hello'), model.Event('hello'))
        with self.assertRaises(TypeError):
            model.Event()

    def test_with_parameters(self):
        event = model.Event('test', a=1, b=2, c=3)
        self.assertEqual(event.data, {'a': 1, 'b': 2, 'c': 3})

    def test_parameter_name(self):
        with self.assertRaises(TypeError):
            model.Event('test', name='fail')

    def test_parameters_access(self):
        event = model.Event('test', a=1, b=2, c=3)
        self.assertEqual(event.a, 1)
        self.assertEqual(event.b, 2)
        self.assertEqual(event.c, 3)

        with self.assertRaises(AttributeError):
            _ = event.d

        with self.assertRaises(KeyError):
            _ = event.data['d']


class TraversalTests(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/composite.yaml') as f:
            self.sc = io.import_from_yaml(f)

    def test_parent(self):
        self.assertEqual(self.sc.parent_for('s2'), 'root')
        self.assertEqual(self.sc.parent_for('root'), None)
        with self.assertRaises(exceptions.StatechartError):
            self.sc.parent_for('unknown')

    def test_children(self):
        self.assertEqual(sorted(self.sc.children_for('root')), ['s1', 's2'])
        self.assertEqual(self.sc.children_for('s1b2'), [])

    def test_ancestors(self):
        self.assertEqual(self.sc.ancestors_for('s2'), ['root'])
        self.assertEqual(self.sc.ancestors_for('s1a'), ['s1', 'root'])
        self.assertEqual(self.sc.ancestors_for('s1b1'), ['s1b', 's1', 'root'])

    def test_descendants(self):
        self.assertEqual(self.sc.descendants_for('s2'), [])
        s1 = self.sc.descendants_for('s1')
        self.assertTrue('s1a' in s1)
        self.assertTrue('s1b' in s1)
        self.assertTrue('s1b1' in s1)
        self.assertTrue('s1b2' in s1)
        self.assertTrue(len(s1) == 4)

    def test_depth(self):
        self.assertEqual(self.sc.depth_for('root'), 1)
        self.assertEqual(self.sc.depth_for('s1'), 2)
        self.assertEqual(self.sc.depth_for('s1a'), 3)
        self.assertEqual(self.sc.depth_for('s1b1'), 4)
        self.assertEqual(self.sc.depth_for('s2'), 2)
        self.assertEqual(self.sc.depth_for('s1b2'), 4)

    def test_lca(self):
        self.assertEqual(self.sc.least_common_ancestor('s1', 's2'), 'root')
        self.assertEqual(self.sc.least_common_ancestor('s1', 's1a'), 'root')
        self.assertEqual(self.sc.least_common_ancestor('s1a', 's1b'), 's1')
        self.assertEqual(self.sc.least_common_ancestor('s1a', 's1b1'), 's1')

    def test_leaf(self):
        self.assertEqual(sorted(self.sc.leaf_for([])), [])
        self.assertEqual(sorted(self.sc.leaf_for(['s1'])), ['s1'])
        self.assertEqual(sorted(self.sc.leaf_for(['s1', 's2'])), ['s1', 's2'])
        self.assertEqual(sorted(self.sc.leaf_for(['s1', 's1b1', 's2'])), ['s1b1', 's2'])
        self.assertEqual(sorted(self.sc.leaf_for(['s1', 's1b', 's1b1'])), ['s1b1'])

    def test_events(self):
        self.assertEqual(self.sc.events_for(), ['click', 'close', 'validate'])
        self.assertEqual(self.sc.events_for('s1b1'), ['validate'])
        self.assertEqual(self.sc.events_for(['s1b1', 's1b']), ['click', 'validate'])

    def test_name_collision(self):
        root = model.CompoundState('root', 'a')
        sc = model.Statechart('test')
        sc.add_state(root, None)
        s1 = model.BasicState('a')
        s2 = model.BasicState('a')
        sc.add_state(s1, parent='root')
        with self.assertRaises(exceptions.StatechartError) as cm:
            sc.add_state(s2, parent='root')
        self.assertIn('already exists!', str(cm.exception))

    def test_root_already_defined(self):
        root = model.CompoundState('root', 'a')
        sc = model.Statechart('test')
        sc.add_state(root, None)
        with self.assertRaises(exceptions.StatechartError) as cm:
            sc.add_state(root, None)
        self.assertIn('already exists!', str(cm.exception))


class ValidateTests(unittest.TestCase):
    # REMARK: "Positive" tests are already done during io.import_from_yaml!
    def test_history_memory_not_child(self):
        with open('tests/yaml/history.yaml') as f:
            statechart = io.import_from_yaml(f)

        statechart.state_for('loop.H').memory = 'pause'
        with self.assertRaises(exceptions.StatechartError) as cm:
            statechart._validate_historystate_memory()
        self.assertIn('Initial memory', str(cm.exception))
        self.assertIn('must be a parent\'s child', str(cm.exception))

    def test_history_memory_unknown(self):
        with open('tests/yaml/history.yaml') as f:
            statechart = io.import_from_yaml(f)

        statechart.state_for('loop.H').memory = 'unknown'
        with self.assertRaises(exceptions.StatechartError) as cm:
            statechart._validate_historystate_memory()
        self.assertIn('Initial memory', str(cm.exception))
        self.assertIn('does not exist', str(cm.exception))

    def test_history_memory_self(self):
        with open('tests/yaml/history.yaml') as f:
            statechart = io.import_from_yaml(f)

        statechart.state_for('loop.H').memory = 'loop.H'
        with self.assertRaises(exceptions.StatechartError) as cm:
            statechart._validate_historystate_memory()
        self.assertIn('Initial memory', str(cm.exception))
        self.assertIn('cannot target itself', str(cm.exception))

    def test_compound_initial_unknown(self):
        with open('tests/yaml/composite.yaml') as f:
            statechart = io.import_from_yaml(f)

        statechart.state_for('s1b').initial = 'unknown'
        with self.assertRaises(exceptions.StatechartError) as cm:
            statechart._validate_compoundstate_initial()
        self.assertIn('Initial state', str(cm.exception))
        self.assertIn('does not exist', str(cm.exception))

    def test_compound_initial_not_child(self):
        with open('tests/yaml/composite.yaml') as f:
            statechart = io.import_from_yaml(f)

        statechart.state_for('s1b').initial = 's1'
        with self.assertRaises(exceptions.StatechartError) as cm:
            statechart._validate_compoundstate_initial()
        self.assertIn('Initial state', str(cm.exception))
        self.assertIn('must be a child state', str(cm.exception))


class TransitionsTests(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/internal.yaml') as f:
            self.sc = io.import_from_yaml(f)

    def test_transitions_from(self):
        self.assertEqual(self.sc.transitions_from('root'), [])
        self.assertEqual(len(self.sc.transitions_from('active')), 2)
        self.assertEqual(len(self.sc.transitions_from('s1')), 1)

        nb_transitions = 0
        for transition in self.sc.transitions:
            if transition.source == 'unknown':
                nb_transitions += 1
        self.assertEqual(nb_transitions, 0)

        for transition in self.sc.transitions_from('active'):
            self.assertEqual(transition.source, 'active')

    def test_transitions_to(self):
        self.assertEqual(self.sc.transitions_to('root'), [])
        self.assertEqual(len(self.sc.transitions_to('s1')), 1)
        self.assertEqual(len(self.sc.transitions_to('s2')), 2)

        nb_transitions = 0
        for transition in self.sc.transitions:
            if transition.source == 'unknown':
                nb_transitions += 1
        self.assertEqual(nb_transitions, 0)

        for transition in self.sc.transitions_from('s2'):
            self.assertEqual(transition.target, 's2')

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.add_transition(model.Transition('s2'))
        self.assertIn('Cannot add', str(cm.exception))

        self.sc.add_transition(model.Transition('s1'))
        self.assertEqual(len(self.sc.transitions_to('s1')), 2)

    def test_transitions_with(self):
        self.assertEqual(len(self.sc.transitions_with('next')), 1)
        self.assertEqual(len(self.sc.transitions_with('unknown')), 0)


class TransitionRotationTests(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/internal.yaml') as f:
            self.sc = io.import_from_yaml(f)

    def test_rotate_source(self):
        tr = next(t for t in self.sc.transitions if t.source == 's1')
        self.sc.rotate_transition(tr, new_source='s1')
        self.assertEqual(tr.source, 's1')

        self.sc.rotate_transition(tr, new_source='active')
        self.assertEqual(tr.source, 'active')

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.rotate_transition(tr, new_source=None)
        self.assertIn('State', str(cm.exception))
        self.assertIn('does not exist', str(cm.exception))

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.rotate_transition(tr, new_source='s2')
        self.assertIn('cannot have transitions', str(cm.exception))

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.rotate_transition(tr, new_source='unknown')
        self.assertIn('State', str(cm.exception))
        self.assertIn('does not exist', str(cm.exception))

        self.sc.validate()

    def test_rotate_target(self):
        tr = next(t for t in self.sc.transitions if t.source == 's1')
        self.sc.rotate_transition(tr, new_target='s2')
        self.assertEqual(tr.target, 's2')

        self.sc.rotate_transition(tr, new_target='active')
        self.assertEqual(tr.target, 'active')

        self.sc.rotate_transition(tr, new_target=None)
        self.assertEqual(tr.target, None)
        self.assertTrue(tr.internal)

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.rotate_transition(tr, new_target='unknown')
        self.assertIn('State', str(cm.exception))
        self.assertIn('does not exist', str(cm.exception))

        self.sc.validate()

    def test_rotate_both(self):
        tr = next(t for t in self.sc.transitions if t.source == 's1')

        self.sc.rotate_transition(tr, new_source='s1', new_target='s2')
        self.assertEqual(tr.source, 's1')
        self.assertEqual(tr.target, 's2')
        self.sc.validate()

    def test_rotate_both_unexisting(self):
        tr = next(t for t in self.sc.transitions if t.source == 's1')

        with self.assertRaises(ValueError):
            self.sc.rotate_transition(tr)

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.rotate_transition(tr, new_source=None, new_target=None)
        self.assertIn('State', str(cm.exception))
        self.assertIn('does not exist', str(cm.exception))

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.rotate_transition(tr, new_source='s2', new_target='s2')
        self.assertIn('State', str(cm.exception))
        self.assertIn('cannot have transitions', str(cm.exception))

        self.sc.validate()

    def test_rotate_both_with_internal(self):
        tr = next(t for t in self.sc.transitions if t.source == 's1')

        self.sc.rotate_transition(tr, new_source='s1', new_target=None)
        self.assertEqual(tr.source, 's1')
        self.assertEqual(tr.target, None)
        self.assertTrue(tr.internal)
        self.sc.validate()


class RemoveTransitionsTests(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/internal.yaml') as f:
            self.sc = io.import_from_yaml(f)

    def test_remove_existing_transition(self):
        transitions = self.sc.transitions
        for transition in transitions:
            self.sc.remove_transition(transition)
        self.assertEqual(len(self.sc.transitions), 0)
        self.sc.validate()

    def test_remove_unexisting_transition(self):
        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.remove_transition(None)
        self.assertIn('Transition', str(cm.exception))
        self.assertIn('does not exist', str(cm.exception))

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.remove_transition(model.Transition('a', 'b'))
        self.assertIn('Transition', str(cm.exception))
        self.assertIn('does not exist', str(cm.exception))

        self.sc.validate()


class RemoveStatesTests(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/internal.yaml') as f:
            self.sc = io.import_from_yaml(f)

    def test_remove_existing_state(self):
        self.sc.remove_state('active')

        self.assertTrue('active' not in self.sc.states)
        nb_transitions = 0
        for transition in self.sc.transitions:
            if transition.source == 'active':
                nb_transitions += 1
        self.assertEqual(nb_transitions, 0)

        nb_transitions = 0
        for transition in self.sc.transitions:
            if transition.target == 'active':
                nb_transitions += 1
        self.assertEqual(nb_transitions, 0)
        self.sc.validate()

    def test_remove_unregister_parent_children(self):
        self.sc.remove_state('s1')
        self.assertFalse('s1' in self.sc.children_for('active'))
        self.sc.validate()

    def test_remove_unexisting_state(self):
        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.remove_state('unknown')
        self.assertIn('State', str(cm.exception))
        self.assertIn('does not exist', str(cm.exception))

        self.sc.validate()

    def test_remove_root_state(self):
        self.sc.remove_state('root')
        self.assertEqual(len(self.sc.transitions), 0)
        self.assertEqual(len(self.sc.states), 0)
        self.assertEqual(self.sc.root, None)
        self.sc.validate()

    def test_remove_nested_states(self):
        with open('tests/yaml/composite.yaml') as f:
            self.sc = io.import_from_yaml(f)
        self.sc.remove_state('s1')
        self.assertFalse('s1a' in self.sc.states)
        self.sc.validate()

        with open('tests/yaml/composite.yaml') as f:
            self.sc = io.import_from_yaml(f)
        self.sc.remove_state('s1a')
        self.assertFalse('s1a' in self.sc.states)
        self.sc.validate()


class RenameStatesTests(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/internal.yaml') as f:
            self.sc = io.import_from_yaml(f)

    def test_rename_unexisting_state(self):
        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.rename_state('unknown', 's3')
        self.assertIn('State', str(cm.exception))
        self.assertIn('does not exist', str(cm.exception))

        self.sc.validate()

    def test_do_not_change_name(self):
        self.sc.rename_state('s2', 's2')

        self.assertTrue('s2' in self.sc.states)
        self.assertEqual(self.sc.parent_for('s2'), 'root')
        self.assertTrue('s2' in self.sc.children_for('root'))

        self.sc.validate()

    def test_rename_to_an_existing_state(self):
        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.rename_state('s2', 's1')
        self.assertIn('State', str(cm.exception))
        self.assertIn('already exists!', str(cm.exception))

        self.sc.validate()

    def test_rename_old_disappears(self):
        with open('tests/yaml/composite.yaml') as f:
            self.sc = io.import_from_yaml(f)
        self.sc.rename_state('s1', 'new s1')

        self.assertFalse('s1' in self.sc.states)
        self.assertNotEqual('s1', self.sc.parent_for('s1a'))
        self.assertFalse('s1' in self.sc.children_for('root'))

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.state_for('s1')
        self.assertIn('State', str(cm.exception))
        self.assertIn('does not exist', str(cm.exception))

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.children_for('s1')
        self.assertIn('State', str(cm.exception))
        self.assertIn('does not exist', str(cm.exception))

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.parent_for('s1')
        self.assertIn('State', str(cm.exception))
        self.assertIn('does not exist', str(cm.exception))

        self.sc.validate()

    def test_rename_change_state_name(self):
        with open('tests/yaml/composite.yaml') as f:
            self.sc = io.import_from_yaml(f)

        state = self.sc.state_for('s1')
        self.sc.rename_state('s1', 'new s1')
        self.assertEqual(state.name, 'new s1')

    def test_rename_new_appears(self):
        with open('tests/yaml/composite.yaml') as f:
            self.sc = io.import_from_yaml(f)
        self.sc.rename_state('s1', 'new s1')

        self.assertTrue('new s1' in self.sc.states)
        self.assertEqual('new s1', self.sc.parent_for('s1a'))
        self.assertTrue('new s1' in self.sc.children_for('root'))

        self.sc.state_for('new s1')
        self.sc.children_for('new s1')
        self.sc.parent_for('new s1')
        self.sc.validate()

    def test_rename_adapt_transitions(self):
        self.sc.rename_state('s1', 's3')

        nb_transitions = 0
        for transition in self.sc.transitions:
            if transition.source == 's1':
                nb_transitions += 1
        self.assertEqual(nb_transitions, 0)

        nb_transitions = 0
        for transition in self.sc.transitions:
            if transition.target == 's1':
                nb_transitions += 1
        self.assertEqual(nb_transitions, 0)

        self.assertTrue(len(self.sc.transitions_from('s3')), 1)
        self.assertTrue(len(self.sc.transitions_to('s2')), 1)
        self.sc.validate()

    def test_rename_root(self):
        self.sc.rename_state('root', 'new root')
        self.assertEqual(self.sc.parent_for('new root'), None)
        self.assertEqual(self.sc.parent_for('s1'), 'new root')
        self.assertFalse('root' in self.sc.states)

        self.assertTrue('new root' in self.sc.states)
        self.assertFalse('root' in self.sc.states)
        self.assertEqual(self.sc.parent_for('new root'), None)
        self.assertEqual(self.sc.parent_for('active'), 'new root')
        self.assertTrue('active' in self.sc.children_for('new root'))

        self.sc.validate()

    def test_rename_change_initial(self):
        self.sc.rename_state('active', 's3')
        self.assertEqual(self.sc.state_for('root').initial, 's3')
        self.sc.validate()

    def test_rename_change_memory(self):
        with open('tests/yaml/history.yaml') as f:
            self.sc = io.import_from_yaml(f)
        self.sc.state_for('loop.H').memory = 's1'
        self.sc.rename_state('s1', 's4')
        self.assertEqual(self.sc.state_for('loop.H').memory, 's4')
        self.sc.validate()


class MoveStateTest(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/composite.yaml') as f:
            self.sc = io.import_from_yaml(f)

    def test_move_simple(self):
        self.sc.move_state('s1b2', 's1b1')
        self.assertTrue('s1b2' in self.sc.children_for('s1b1'))
        self.assertFalse('s1b2' in self.sc.children_for('s1b'))
        self.assertEqual('s1b1', self.sc.parent_for('s1b2'))
        self.sc.validate()

    def test_move_composite(self):
        self.sc.move_state('s1b', 's1a')
        self.assertTrue('s1b' in self.sc.children_for('s1a'))
        self.assertFalse('s1b' in self.sc.children_for('s1'))
        self.assertEqual('s1a', self.sc.parent_for('s1b'))
        self.sc.validate()

    def test_move_to_descendant(self):
        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.move_state('s1b', 's1b')
        self.assertIn('State', str(cm.exception))
        self.assertIn('cannot be moved into itself or one of its descendants', str(cm.exception))

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.move_state('s1b', 's1b1')
        self.assertIn('State', str(cm.exception))
        self.assertIn('cannot be moved into itself or one of its descendants', str(cm.exception))

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.move_state('s1', 's1b1')
        self.assertIn('State', str(cm.exception))
        self.assertIn('cannot be moved into itself or one of its descendants', str(cm.exception))

        self.sc.validate()

    def test_move_unexisting(self):
        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.move_state('unknown', 's1b1')
        self.assertIn('State', str(cm.exception))
        self.assertIn('does not exist', str(cm.exception))

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc.move_state('s1b1', 'unknown')
        self.assertIn('State', str(cm.exception))
        self.assertIn('does not exist', str(cm.exception))

        self.sc.validate()

    def test_move_with_initial(self):
        self.sc.move_state('s1a', 'root')
        self.assertEqual(self.sc.state_for('s1').initial, None)
        self.sc.validate()

    def test_move_with_memory(self):
        with open('tests/yaml/history.yaml') as f:
            self.sc = io.import_from_yaml(f)
        self.sc.state_for('loop.H').memory = 's1'
        self.sc.move_state('s1', 's2')
        self.assertEqual(self.sc.state_for('loop.H').memory, None)
        self.sc.validate()


class CopyFromStatechartTests(unittest.TestCase):
    def setUp(self):
        with open('tests/yaml/simple.yaml') as f:
            self.sc1 = io.import_from_yaml(f)
        with open('tests/yaml/composite.yaml') as f:
            self.sc2 = io.import_from_yaml(f)

        for state in self.sc1.states:
            self.sc1.rename_state(state, 'sc1_' + state)

    def test_keep_other_states(self):
        sc1_states = self.sc1.states

        self.sc2.copy_from_statechart(self.sc1, 'sc1_root', replace='s1a')
        self.assertIn('s1a', self.sc2.states)
        self.assertNotIn('sc1_root', self.sc2.states)

        sc1_states.remove('sc1_root')
        for state in sc1_states:
            self.assertIn(state, self.sc2.states)

    def test_keep_other_transitions(self):
        sc1_transitions = self.sc1.transitions

        self.sc2.copy_from_statechart(self.sc1, 'sc1_root', replace='s1b1')

        for transition in sc1_transitions:
            self.assertIn(transition, self.sc2.transitions)

    def test_keep_existing_states(self):
        sc2_states = self.sc2.states

        self.sc2.copy_from_statechart(self.sc1, 'sc1_root', replace='s1b1')

        for state in sc2_states:
            self.assertIn(state, self.sc2.states)

    def test_keep_existing_transitions(self):
        sc2_transitions = self.sc2.transitions

        self.sc2.copy_from_statechart(self.sc1, 'sc1_root', replace='s1b1')

        for transition in sc2_transitions:
            self.assertIn(transition, self.sc2.transitions)

    def test_orphaned_transitions(self):
        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc2.copy_from_statechart(self.sc1, 'sc1_s1', replace='s1b1')
        self.assertIn('not contained in', str(cm.exception))

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc2.copy_from_statechart(self.sc1, 'sc1_final', replace='s1b1')
        self.assertIn('not contained in', str(cm.exception))

    def test_invalid_plug(self):
        # On compound
        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc2.copy_from_statechart(self.sc1, 'sc1_root', replace='s1')
        self.assertIn('children', str(cm.exception))

        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc2.copy_from_statechart(self.sc1, 'sc1_root', replace='s1b')
        self.assertIn('children', str(cm.exception))

    def test_with_namespace(self):
        sc1_states = self.sc1.states
        sc2_states = self.sc2.states

        namespace = lambda s: '__' + s
        self.sc2.copy_from_statechart(self.sc1, 'sc1_root', replace='s1a', renaming_func=namespace)

        expected_states = sc2_states + [namespace(s) for s in sc1_states]
        expected_states.remove(namespace(self.sc1.root))

        self.assertSetEqual(set(expected_states), set(self.sc2.states))

    def test_conflicting_names(self):
        self.sc1.rename_state('sc1_s1', 's1')
        with self.assertRaises(exceptions.StatechartError) as cm:
            self.sc2.copy_from_statechart(self.sc1, 'sc1_root', replace='s1a')
        self.assertIn('already exists', str(cm.exception))



from typing import Iterable

import ruamel.yaml as yaml
import schema
from sismic.exceptions import StatechartError
from sismic.model import Statechart

from .datadict import export_to_dict, import_from_dict

__all__ = ['import_from_yaml', 'export_to_yaml']


class SCHEMA:
    contract = {schema.Or('before', 'after', 'always', 'sequentially'): schema.Use(str)}

    transition = {
        schema.Optional('target'): schema.Use(str),
        schema.Optional('event'): schema.Use(str),
        schema.Optional('guard'): schema.Use(str),
        schema.Optional('action'): schema.Use(str),
        schema.Optional('contract'): [contract],
    }

    state = dict()  # type: Dict
    state.update({
        'name': schema.Use(str),
        schema.Optional('type'): schema.Or('final', 'shallow history', 'deep history'),
        schema.Optional('on entry'): schema.Use(str),
        schema.Optional('on exit'): schema.Use(str),
        schema.Optional('transitions'): [transition],
        schema.Optional('contract'): [contract],
        schema.Optional('initial'): schema.Use(str),
        schema.Optional('parallel states'): [state],
        schema.Optional('states'): [state],
    })

    statechart = {
        'statechart': {
            'name': schema.Use(str),
            schema.Optional('description'): schema.Use(str),
            schema.Optional('preamble'): schema.Use(str),
            'root state': state,
        }
    }


def import_from_yaml(statechart: Iterable[str], ignore_schema: bool=False, ignore_validation: bool=False) -> Statechart:
    """
    Import a statechart from a YAML representation.

    Unless specified, the structure contained in the YAML is validated against a predefined
    schema (see *sismic.io.SCHEMA*), and the resulting statechart is validated using its *validate()* method.

    :param statechart: string or any equivalent object
    :param ignore_schema: set to *True* to disable yaml validation.
    :param ignore_validation: set to *True* to disable statechart validation.
    :return: a *Statechart* instance
    """
    data = yaml.load(statechart, Loader=yaml.BaseLoader)  # type: dict

    if not ignore_schema:
        try:
            data = schema.Schema(SCHEMA.statechart).validate(data)
        except schema.SchemaError as e:
            raise StatechartError('YAML validation failed') from e

    sc = import_from_dict(data)

    if not ignore_validation:
        sc.validate()
    return sc


def export_to_yaml(statechart: Statechart) -> str:
    """
    Export given *Statechart* instance to YAML

    :param statechart:
    :return: A textual YAML representation
    """
    return yaml.dump(export_to_dict(statechart, ordered=False),
                     width=1000, default_flow_style=False, default_style='"')



from behave import given, then, when  # type: ignore
from sismic.interpreter import Interpreter
from sismic.interpreter.helpers import log_trace
from sismic.io import import_from_yaml
from sismic.model import Event


# #################### GENERAL PURPOSE
@given('I do nothing')
@when('I do nothing')
def do_nothing(context):
    pass


@given('I reproduce "{scenario}"')
@when('I reproduce "{scenario}"')
def reproduce_scenario(context, scenario):
    current_feature = context.feature
    for included_scenario in current_feature.scenarios:
        if included_scenario.name == scenario:
            steps = ['{} {}'.format(s.step_type, s.name) for s in included_scenario.steps]
            context.execute_steps('\n'.join(steps))
            return
    assert False, 'Unknown scenario {}.'.format(scenario)


@given('I repeat step "{step}" {repeat:d} times')
@when('I repeat step "{step}" {repeat:d} times')
def repeat_step(context, step, repeat):
    keyword = step.split(' ', 1)[0].lower()
    assert keyword in ['given', 'when', 'and', 'but', 'then'], \
        'Step {} should start with a supported keyword'.format(step)

    for _ in range(repeat):
        context.execute_steps(step)


# #################### CONFIGURATION
def _execute_statechart(context, force_execution=False, execute_once=False):
    if context._automatic_execution or force_execution:
        # Remove events
        while len(context._events) > 0:
            context._events.pop()

        if execute_once:
            context._interpreter.execute_once()
        else:
            context._interpreter.execute()


@given('I disable automatic execution')
def disable_automatic_execution(context):
    context._automatic_execution = False


@given('I enable automatic execution')
def enable_automatic_execution(context):
    context._automatic_execution = True


@given('I import a statechart from {path}')
@when('I import a statechart from {path}')
def load_statechart(context, path):
    with open(path) as f:
        context._statechart = import_from_yaml(f)
    context._interpreter = Interpreter(context._statechart)
    context._steps = log_trace(context._interpreter)

    context._automatic_execution = True

    context._events = []
    context._interpreter.bind(context._events.append)

    _execute_statechart(context, force_execution=True, execute_once=True)


@given('I execute the statechart')
@when('I execute the statechart')
def execute_statechart(context):
    _execute_statechart(context, force_execution=True)


@given('I execute once the statechart')
@when('I execute once the statechart')
def execute_once_statechart(context):
    _execute_statechart(context, force_execution=True, execute_once=True)


# #################### STATECHART
@given('I send event {event_name}')
@given('I send event {event_name} with {parameter}={value}')
@when('I send event {event_name}')
@when('I send event {event_name} with {parameter}={value}')
def send_event(context, event_name, parameter=None, value=None):
    parameters = {}
    if context.table:
        for row in context.table:
            parameters[row['parameter']] = eval(row['value'], {}, {})

    if parameter and value:
        parameters[parameter] = eval(value)

    event = Event(event_name, **parameters)
    context._interpreter.queue(event)
    _execute_statechart(context)


@given('I wait {seconds:g} seconds')
@given('I wait {seconds:g} second')
@when('I wait {seconds:g} seconds')
@when('I wait {seconds:g} second')
def wait_seconds_once(context, seconds):
    context._interpreter.time += seconds
    _execute_statechart(context)


@given('I wait {seconds:g} seconds {repeat:d} times')
@given('I wait {seconds:g} second {repeat:d} times')
@when('I wait {seconds:g} seconds {repeat:d} times')
@when('I wait {seconds:g} second {repeat:d} times')
def wait_seconds(context, seconds, repeat):
    for _ in range(repeat):
        wait_seconds_once(context, seconds)


@given('I set variable {variable_name} to {value}')
def set_variable(context, variable_name, value):
    context._interpreter.context[variable_name] = eval(value, {}, {})


@then('state {state_name} should be active')
def state_is_active(context, state_name):
    assert state_name in context._statechart.states, 'Unknown state {}'.format(state_name)
    assert state_name in context._interpreter.configuration, 'State {} is not active'.format(state_name)


@then('state {state_name} should not be active')
def state_is_not_active(context, state_name):
    assert state_name in context._statechart.states, 'Unknown state {}'.format(state_name)
    assert state_name not in context._interpreter.configuration, 'State {} is active'.format(state_name)


@then('event {event_name} should be fired')
@then('event {event_name} should be fired with {parameter}={value}')
def event_is_received(context, event_name, parameter=None, value=None):
    parameters = {}
    if context.table:
        for row in context.table:
            parameters[row['parameter']] = eval(row['value'], {}, {})

    if parameter and value:
        parameters[parameter] = eval(value)

    for event in context._events:
        if event.name == event_name:
            matching_parameters = True
            for key, value in parameters.items():
                if getattr(event, key, None) != value:
                    matching_parameters = False
                    break
            if matching_parameters:
                return

    assert False, 'No matching event fired for {} with {} in {}'.format(event_name, parameters, context._events)


@then('event {event_name} should not be fired')
def event_is_not_received(context, event_name):
    for event in context._events:
        if event.name == event_name:
            assert False, 'Event {} was raised'.format(event)


@then('no event should be fired')
def no_event_received(context):
    assert len(context._events) == 0, 'Sent events: {}'.format(context._events)


@then('variable {variable_name} should be defined')
def variable_is_defined(context, variable_name):
    assert variable_name in context._interpreter.context, '{} is not defined'.format(variable_name)


@then('the value of {variable_name} should be {value}')
def variable_equals_value(context, variable_name, value):
    variable_is_defined(context, variable_name)
    value = eval(value, {}, {})
    context_value = context._interpreter.context[variable_name]
    assert context_value == value, 'Variable {} equals {} != {}'.format(variable_name, context_value, value)


@then('expression {expression} should hold')
def expression_holds(context, expression):
    assert context._interpreter._evaluator._evaluate_code(expression), '{} does not hold'.format(expression)


@then('the statechart is in a final configuration')
def final_configuration(context):
    assert context._interpreter.final, \
        'The statechart is not in a final configuration: {}'.format(context._interpreter.configuration)
